#!/usr/bin/env python
# -*- coding: utf-8 -*-
# (c) Copyright IBM Corp. 2010, 2022. All Rights Reserved

# resilient_common logging constants
WARN_BUILD_INCIDENT_ORG_ID = "Building an incident link without the ORG ID may produce a link that directs the user to a different incident resulting in unexpected results. It is recommended to provide the ORG ID." # pragma: no cover

MIN_SUPPORTED_PY_VERSION = (3, 6)